// -*- C++ -*-
/*!
 * @file Factory.h
 * @brief generic Factory template class
 * @date $Date$
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2009, 2012
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id$
 *
 */

#ifndef COIL_FACTORY_H
#define COIL_FACTORY_H

#include <assert.h>
#include <string>
#include <map>
#include <algorithm>
#include <vector>
#include <coil/Singleton.h>

// for Windows DLL export
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__)
#   ifdef LIBRARY_EXPORTS
#      define EXTERN
#      define DLL_PLUGIN __declspec(dllexport)
#   else
#      define EXTERN extern
#      define DLL_PLUGIN __declspec(dllimport)
#   endif
#else 
#   define DLL_PLUGIN
#ifndef EXTERN
#   define EXTERN
#endif // ifndef EXTERN
#endif /* Windows */

#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__)
#pragma warning( push )
#pragma warning( disable : 4251 )
#endif

namespace coil
{
  /*!
   * @if jp
   *
   * @brief Creator ev[g
   *
   * @else
   *
   * @brief Creator template
   *
   * @endif
   */
  template <class AbstractClass, class ConcreteClass>
  AbstractClass* Creator()
  {
    return new ConcreteClass();
  }

  /*!
   * @if jp
   *
   * @brief Destructor ev[g
   *
   * @else
   *
   * @brief Destructor template
   *
   * @endif
   */
  template <class AbstractClass, class ConcreteClass>
  void Destructor(AbstractClass*& obj)
  {
    if (obj == 0) { return; }
    ConcreteClass* tmp = dynamic_cast<ConcreteClass*>(obj);
    if (tmp == 0) { return; }
    delete obj;
    obj = 0;
  }

  /*!
   * @if jp
   *
   * @class Factory
   * @brief Factory ev[gNX
   *
   * @else
   *
   * @class Factory
   * @brief Factory template class
   *
   * @endif
   */
  template <
    class AbstractClass,
    typename Identifier = std::string,
    typename Compare = std::less<Identifier>,
    typename Creator = AbstractClass* (*)(),
    typename Destructor = void (*)(AbstractClass*&)
    >
  class Factory
  {
    class FactoryEntry;
  public:

    typedef std::map<Identifier, FactoryEntry> FactoryMap;
    typedef typename FactoryMap::iterator      FactoryMapIt;
    typedef std::map<AbstractClass*, FactoryEntry> ObjectMap;
    typedef typename ObjectMap::iterator           ObjectMapIt;
    
    enum ReturnCode
      {
        FACTORY_OK,
        FACTORY_ERROR,
        ALREADY_EXISTS,
        NOT_FOUND,
        INVALID_ARG,
        UNKNOWN_ERROR
      };

    /*!
     * @if jp
     *
     * @brief t@Ng[L`FbN
     *
     * wID̃t@Ng[LԂB
     *
     * @param id t@Ng[ID
     *
     * @return true: L, false: 
     *
     * @else
     *
     * @brief Factory presence check
     *
     * Return a factory having ID or not.
     *
     * @param id Factory ID.
     *
     * @return true: found, false: not found
     *
     * @endif
     */
    bool hasFactory(const Identifier& id)
    {
      if (m_creators.count(id) == 0) { return false; }
      return true;
    }

    /*!
     * @if jp
     *
     * @brief t@Ng[IDXg擾
     *
     * t@Ng[IDXgԂB
     *
     * @return ID Xg
     *
     * @else
     *
     * @brief Get factory ID list
     *
     * Return a list of factory ID.
     *
     * @return ID list
     *
     * @endif
     */
    std::vector<Identifier> getIdentifiers()
    {
      std::vector<Identifier> idlist;
      idlist.reserve(m_creators.size());

      FactoryMapIt it(m_creators.begin());
      FactoryMapIt it_end(m_creators.end());

      while (it != it_end)
        {
          idlist.push_back(it->first);
          ++it;
        }
      return idlist;
    }

    /*!
     * @if jp
     *
     * @brief t@Ng[o^
     *
     * t@Ng[o^B
     *
     * @param id t@Ng[ID
     * @param creator NG[^pt@N^
     * @param destructor fXgN^pt@N^
     *
     * @return FACTORY_OK: I
     *         ALREADY_EXISTS: o^ς
     *         INVALID_ARG: creator  destructor sȒl܂
     *
     * @else
     *
     * @brief Add factory
     *
     * Add a factory.
     *
     * @param id Factory ID.
     * @param creator Functor for creator.
     * @param destructor Functor for destructor.
     *
     * @return FACTORY_OK: Successful
     *         ALREADY_EXISTS: already exists.
     *         INVALID_ARG: creator or destructor with invalid values.
     *
     * @endif
     */
    ReturnCode addFactory(const Identifier& id,
                          Creator creator,
                          Destructor destructor)
    {
      if (creator == 0 || destructor == 0) { return INVALID_ARG; }
      if (m_creators.count(id) != 0) { return ALREADY_EXISTS; }
      FactoryEntry f(id, creator, destructor);
      m_creators[id] = f;
      return FACTORY_OK;
    }

    /*!
     * @if jp
     *
     * @brief t@Ng[폜
     *
     * t@Ng[폜B
     *
     * @param id t@Ng[ID
     *
     * @return FACTORY_OK: I
     *         NOT_FOUND: IDo^
     *
     * @else
     *
     * @brief Remove factory
     *
     * Remove a factory.
     *
     * @param id Factory ID.
     *
     * @return FACTORY_OK: Successful
     *         NOT_FOUND: ID not found
     *
     * @endif
     */
    ReturnCode removeFactory(const Identifier& id)
    {
      if (m_creators.count(id) == 0) { return NOT_FOUND; }
      m_creators.erase(id);
      return FACTORY_OK;
    }

    /*!
     * @if jp
     *
     * @brief t@Ng[IuWFNg
     *
     * t@Ng[IuWFNg𐶐B
     *
     * @param id t@Ng[ID
     *
     * @return AbstractClass ̃|C^
     *
     * @else
     *
     * @brief Create factory object
     *
     * Create a factory object.
     *
     * @param id Factory ID.
     *
     * @return Pointer of AbstractClass
     *
     * @endif
     */
    AbstractClass* createObject(const Identifier& id)
    {
      if (m_creators.count(id) == 0) { return 0; }
      AbstractClass* obj = m_creators[id].creator_();
      assert(m_objects.count(obj) == 0);
      m_objects[obj] = m_creators[id];
      return obj;
    }

    /*!
     * @if jp
     *
     * @brief t@Ng[IuWFNg폜
     *
     * t@Ng[IuWFNg폜B
     *
     * @param id t@Ng[ID
     * @param obj t@Ng[IuWFNg
     *
     * @else
     *
     * @brief Delete factory object
     *
     * Delete a factory object.
     *
     * @param id Factory ID.
     * @param obj Factory object.
     *
     * @endif
     */
    ReturnCode deleteObject(const Identifier& id, AbstractClass*& obj)
    {
      if (m_creators.count(id) == 0)
        {
          return deleteObject(obj);
        }
      m_creators[id].destructor_(obj);
      m_objects.erase(obj);
      return FACTORY_OK;
    }

    /*!
     * @if jp
     *
     * @brief t@Ng[IuWFNg폜
     *
     * t@Ng[IuWFNg폜B
     *
     * @param obj t@Ng[IuWFNg
     *
     * @else
     *
     * @brief Delete factory object
     *
     * Delete a factory object.
     *
     * @param obj Factory object.
     *
     * @endif
     */
    ReturnCode deleteObject(AbstractClass*& obj)
    {
      if (m_objects.count(obj) == 0) { return NOT_FOUND; }
      AbstractClass* tmp(obj);
      m_objects[obj].destructor_(obj);
      m_objects.erase(tmp);
      return FACTORY_OK;
    }

    /*!
     * @if jp
     *
     * @brief ς݃IuWFNgXg̎擾
     *
     * ̃t@NgŐꂽIuWFNg̃Xg擾B
     *
     * @return ς݃IuWFNgXg
     *
     * @else
     *
     * @brief Getting created objects
     *
     * This operation returns a list of created objects by the factory.
     *
     * @return created object list
     *
     * @endif
     */
    std::vector<AbstractClass*> createdObjects()
    {
      std::vector<AbstractClass*> objects;
      for (ObjectMapIt it(m_objects.begin()); it != m_objects.end(); ++it)
        {
          objects.push_back(it->first);
        }
      return objects;
    }

    /*!
     * @if jp
     *
     * @brief IuWFNg̃t@Ng̐ǂׂ
     *
     * @param obj ΏۃIuWFNg
     * @return true: ̃t@Ng̐
     *         false: ̃t@Ng̐ł͂Ȃ
     *
     * @else
     *
     * @brief Whether a object is a product of this factory
     *
     * @param obj A target object
     * @return true: The object is a product of the factory
     *         false: The object is not a product of the factory
     *
     * @return created object list
     *
     * @endif
     */
    bool isProducerOf(AbstractClass* obj)
    {
      return m_objects.count(obj) != 0;
    }

    /*!
     * @if jp
     *
     * @brief IuWFNgNXʎq(ID)擾
     *
     * YIuWFNg̃NXʎq(ID)擾B
     *
     * @param obj [in] NXʎq(ID)擾IuWFNg
     * @param id [out] NXʎq(ID)
     * @return ^[R[h NOT_FOUND: ʎq݂Ȃ
     *                        FACTORY_OK: I
     * @else
     *
     * @brief Getting class identifier (ID) from a object
     *
     * This operation returns a class identifier (ID) from a object.
     *
     * @param obj [in] An object to investigate its class ID.
     * @param id [out] Class identifier (ID)
     * @return Return code NOT_FOUND: ID not found
     *                        FACTORY_OK: normal return
     * @endif
     */
    ReturnCode objectToIdentifier(AbstractClass* obj, Identifier& id)
    {
      if (m_objects.count(obj) == 0) { return NOT_FOUND; }
      id = m_objects[obj].id_;
      return FACTORY_OK;
    }

    /*!
     * @if jp
     *
     * @brief IuWFNg̃RXgN^擾
     *
     * ̃t@NgŐꂽIuWFNg̃RXgN^擾B
     * obj ͂̃t@NgŐꂽ̂łȂ΂ȂȂB\
     * isProducerOf() ֐œYIuWFNg̃t@Ng̐
     * 邩ǂ`FbNȂ΂ȂȂB
     *
     * @return IuWFNg̃fXgN^
     *
     * @else
     *
     * @brief Getting destructor of the object
     *
     * This operation returns a constructor of the object created by
     * the factory.  obj must be a product of the factory.  User must
     * check if the object is a product of the factory by using
     * isProducerOf()-function, before using this function.
     *
     * @return destructor of the object
     *
     * @endif
     */
    Creator objectToCreator(AbstractClass* obj)
    {
      return m_objects[obj].creator_;
    }

    /*!
     * @if jp
     *
     * @brief IuWFNg̃fXgN^擾
     *
     * ̃t@NgŐꂽIuWFNg̃fXgN^擾B
     * obj ͂̃t@NgŐꂽ̂łȂ΂ȂȂB\
     * isProducerOf() ֐œYIuWFNg̃t@Ng̐
     * 邩ǂ`FbNȂ΂ȂȂB
     *
     * @return IuWFNg̃fXgN^
     *
     * @else
     *
     * @brief Getting destructor of the object
     *
     * This operation returns a destructor of the object created by
     * the factory.  obj must be a product of the factory.  User must
     * check if the object is a product of the factory by using
     * isProducerOf()-function, before using this function.
     *
     * @return destructor of the object
     *
     * @endif
     */
    Destructor objectToDestructor(AbstractClass* obj)
    {
      return m_objects[obj].destructor_;
    }

  private:

    /*!
     * @if jp
     *
     * @class FactoryEntry
     * @brief FactoryEntry NX
     *
     * @else
     *
     * @class FactoryEntry
     * @brief FactoryEntry class
     *
     * @endif
     */
    class FactoryEntry
    {
    public:
      explicit FactoryEntry()
      {
      }

      /*!
       * @if jp
       *
       * @brief RXgN^
       *
       * RXgN^B
       *
       * @param creator NG[^pt@N^
       * @param destructor fXgN^pt@N^
       *
       * @else
       *
       * @brief Constructor
       *
       * Constructor
       *
       * @param creator Functor for creator.
       * @param destructor Functor for destructor.
       *
       * @endif
       */
      FactoryEntry(Identifier id, Creator creator, Destructor destructor)
        : id_(id), creator_(creator), destructor_(destructor)
      {
      }
      std::string id_;
      Creator creator_;
      Destructor destructor_;
    };
    FactoryMap m_creators;
    ObjectMap  m_objects;
  };



  /*!
   * @if jp
   *
   * @class GlobalFactory
   * @brief GlobalFactory ev[gNX
   *
   * @else
   *
   * @class GlobalFactory
   * @brief GlobalFactory template class
   *
   * @endif
   */
  template <
    class AbstractClass,
    typename Identifier = std::string,
    typename Compare = std::less<Identifier>,
    typename Creator = AbstractClass* (*)(),
    typename Destructor = void (*)(AbstractClass*&)
    >
  class GlobalFactory
    : public Factory<AbstractClass, Identifier, Compare, Creator, Destructor>,
      public coil::Singleton<GlobalFactory<AbstractClass,
                                           Identifier,
                                           Compare,
                                           Creator,
                                           Destructor> >
  {
  public:

  private:
    /*!
     * @if jp
     *
     * @brief RXgN^
     *
     * RXgN^B
     *
     * @else
     *
     * @brief Constructor
     *
     * Constructor
     *
     * @endif
     */
    GlobalFactory(){}

    /*!
     * @if jp
     *
     * @brief fXgN^
     *
     * fXgN^B
     *
     * @else
     *
     * @brief Destructor
     *
     * Destructor
     *
     * @endif
     */
    ~GlobalFactory(){}

    friend class Singleton<GlobalFactory>;
  };

}; // namespace coil


#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__)
#pragma warning( pop )
#endif

#endif // COIL_FACTORY_H
