// -*- C++ -*-
/*!
 * @file ClockManager.h
 * @brief Global clock management class
 * @date $Date$
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2012
 *     Noriaki Ando
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id$
 *
 */

#ifndef COIL_CLOCKMANAGER_H
#define COIL_CLOCKMANAGER_H

#include <string>
#include <coil/Singleton.h>
#include <coil/TimeValue.h>
#include <coil/Mutex.h>
#include <coil/Guard.h>

namespace coil
{
  /*!
   * @if jp
   * @class ꡦ֥ȤΥ󥿡ե
   *
   * Υ饹 ClockManager ˤäƴ륯å֥Ȥ
   * Υ󥿡եǤ롣ClockManager ʣΥå֥
   * Ȥɬפ˱Ŭڤʥå֥Ȥ IClock 
   * եĥ֥ȤȤ֤å֥Ȥñ
   * ˥ƥ֤Τ䡢ȼĥå֥
   * ͤ롣
   *
   * @else
   * @brief An interface to set and get time
   *
   * This class is a interface for clock objects managed by
   * ClockManager. ClockManager manages one or more clocks, and it
   * returns appropriate clock objects according to demands. The clock
   * object might be clock which just returns system time, or a clock
   * which returns individual logical time.
   *
   * @endif
   */
  class IClock
  {
  public:
    virtual ~IClock() {}
    /*!
     * @if jp
     * @brief 
     * @return ߤλ
     * @else
     * @brief Getting time
     * @return Current time
     * @endif
     */
    virtual coil::TimeValue gettime() const = 0;
    /*!
     * @if jp
     * @brief ꤹ
     * @param clocktime ߤλ
     * @else
     * @brief Setting time
     * @param clocktime Current time
     * @endif
     */
    virtual bool settime(coil::TimeValue clocktime) = 0;
  };

  /*!
   * @if jp
   * @class ƥ򰷤å֥
   *
   * Υ饹ϥƥ९åޤϼ륯饹Ǥ롣
   *
   * @else
   * @brief clock object to handle system clock
   *
   * This class sets and gets system clock.
   *
   * @endif
   */
  class SystemClock
    : public IClock
  {
  public:
    virtual ~SystemClock();
    virtual coil::TimeValue gettime() const;
    virtual bool settime(coil::TimeValue clocktime);
  };

  /*!
   * @if jp
   * @class ֤򰷤å֥
   *
   * Υ饹֤ޤϼ륯饹Ǥ롣
   * ñ settime() ˤäꤵ줿 gettime() ˤäƼ뤹
   *
   * @else
   * @brief Clock object to handle logical clock
   *
   * This class sets and gets system clock.
   * It just sets time by settime() and gets time by gettime().
   *
   * @endif
   */
  class LogicalClock
    : public IClock
  {
  public:
    LogicalClock();
    virtual ~LogicalClock();
    virtual coil::TimeValue gettime() const;
    virtual bool settime(coil::TimeValue clocktime);
  private:
    coil::TimeValue m_currentTime;
    mutable coil::Mutex m_currentTimeMutex;
  };

  /*!
   * @if jp
   * @class ĴѤ߻򰷤å֥
   *
   * settime() ƤӽФ˸߻Ȥκ򥪥եåȤȤݻ
   * gettime() ˤäƥեåĴѤߤλ֤
   *
   * @else
   * @brief Clock object to handle adjusted clock
   *
   * This class stores a offset time with current system clock when
   * settime(), and gettime() returns adjusted clock by the offset.
   *
   * @endif
   */
  class AdjustedClock
    : public IClock
  {
  public:
    AdjustedClock();
    virtual ~AdjustedClock();
    virtual coil::TimeValue gettime() const;
    virtual bool settime(coil::TimeValue clocktime);
  private:
    coil::TimeValue m_offset;
    mutable coil::Mutex m_offsetMutex;
  };

  /*!
   * @if jp
   * @class Хʥå饹
   *
   * Υ饹ϥХ˥å֥Ȥ󶡤륷󥰥ȥ
   * 饹Ǥ롣getClocK(å̾) ˤ IClock Υå
   * Ȥ֤Ѳǽʥå "system", "logical" 
   * "adjusted" ΣǤ롣
   *
   * @else
   * @brief A global clock management class
   *
   * This class is a singleton class that provides clock objects
   * globally. It provides a IClock object by getClock(<clock
   * type>). As clock types, "system", "logical" and "adjusted" are
   * available.
   *
   * @endif
   */
  class ClockManager
#ifndef WIN32
    : public coil::Singleton<ClockManager>
#endif
  {
  public:
    IClock& getClock(std::string clocktype);
#ifdef WIN32
    static ClockManager& instance();
    static ClockManager* clockmgr;
    static coil::Mutex clockmgr_mutex;
  private:
	ClockManager() {};
#endif
  private:
    SystemClock   m_systemClock;
    LogicalClock  m_logicalClock;
    AdjustedClock m_adjustedClock;
  };
}; // namespace coil
#endif // COIL_CLOCKMANAGER_H
